/*
 * Crystal Light Code
 *
 * Display mode, direction and speed of display controllable
 * via a webpage.
 * 
 * Hardware:
 *   Esp8266 D1 Mini
 *   A single NeoPixel
 *
 * Concept, design and implementaton by: Craig A. Lindley
 * Last Update: 07/02/2023
 */

#include <ESP8266WiFi.h>
#include <ESPAsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include "WebPage.h"
#include <Adafruit_NeoPixel.h>

// WiFi parameters
#define WIFI_SSID "CraigNet"
#define WIFI_PSWD "craigandheather"
#define WIFI_HOSTNAME "CrystalLight"
#define WIFI_TIMEOUT (5 * 60000)  // Checks WiFi every 5 minutes. Reboot if connection down

// Set web server port number to 80
AsyncWebServer server(80);
AsyncWebSocket ws("/ws");

// Number of NeoPixels is 1
#define NUMBER_OF_PIXELS 1
#define PIXEL_PIN D4
#define BRIGHTNESS 128

// Instantiate NeoPixel driver
Adafruit_NeoPixel pixels(NUMBER_OF_PIXELS, PIXEL_PIN, NEO_GRB + NEO_KHZ800);

// Number of steps in HSV color space between pure colors
#define COLOR_STEP_INC 45

// Speed delay values
#define MEDIUM_DELAY_MS 4000
#define SLOW_DELAY_MS (MEDIUM_DELAY_MS * 4)
#define FAST_DELAY_MS (MEDIUM_DELAY_MS / 4)

// Current operational parameters
uint16_t currentPower;
uint16_t currentMode;
uint16_t currentDir;
uint16_t currentSpeed;
uint16_t currentSteppedDelay;
uint16_t currentSweptedDelay;

bool exitFlag;

uint32_t last_wifi_check_time = 0;

/*********************************************************************/
/***                          Misc Functions                       ***/
/*********************************************************************/

// Check for Internet connectivity by pinging google DNS
// Returns true if Internet is accessible
boolean checkInternetConnection() {
  IPAddress serverIP;

  int result = WiFi.hostByName("8.8.8.8", serverIP);
  return (result == 1);
}

void handleWebSocketMessage(void *arg, uint8_t *data, size_t len) {
  char cmdToken[5];
  char argToken[5];

  AwsFrameInfo *info = (AwsFrameInfo *)arg;
  if (info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT) {
    data[len] = 0;
    memset(cmdToken, 0, sizeof(cmdToken));
    memset(argToken, 0, sizeof(argToken));

    strncpy(cmdToken, (char *)data, 2);
    strncpy(argToken, ((char *)data) + 3, 2);

    // Convert numeric argument to an int
    uint16_t val = atoi(argToken);

    // Process the cmd message
    // Was it a power change message ?
    if (strcmp(cmdToken, "PC") == 0) {
      // Yes
      currentPower = val;
      exitFlag = true;
      // Was it a mode change message ?
    } else if (strcmp(cmdToken, "MC") == 0) {
      // Yes
      currentMode = val;
      exitFlag = true;
      // Was it a direction change message ?
    } else if (strcmp(cmdToken, "DC") == 0) {
      // Yes
      currentDir = val;
      exitFlag = true;
      // Was it a speed change message ?
    } else if (strcmp(cmdToken, "SC") == 0) {
      // Yes
      currentSpeed = val;
      exitFlag = true;
    }
  }
}

void onEvent(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type,
             void *arg, uint8_t *data, size_t len) {
  switch (type) {
    case WS_EVT_CONNECT:
      Serial.printf("WebSocket client #%u connected from %s\n", client->id(), client->remoteIP().toString().c_str());
      break;
    case WS_EVT_DISCONNECT:
      Serial.printf("WebSocket client #%u disconnected\n", client->id());
      break;
    case WS_EVT_DATA:
      handleWebSocketMessage(arg, data, len);
      break;
    case WS_EVT_PONG:
    case WS_EVT_ERROR:
      break;
  }
}

void initWebSocket() {
  ws.onEvent(onEvent);
  server.addHandler(&ws);
}

void clearLED() {
  pixels.clear();
  pixels.show();
}

// Swepted display pattern
void displaySweptedPattern() {
  switch (currentDir) {
    case 0:
      {
        // Swepted up
        for (int i = 0; i < 225; i++) {
          pixels.setPixelColor(0, pixels.ColorHSV(map(i, 0, 255, 0, 65535)));
          pixels.show();
          delay(currentSweptedDelay);
          if (exitFlag) {
            break;
          }
        }
        clearLED();
        delay(1000);
      }
      break;
    case 1:
      {
        // Swepted up then down
        // Swepted up
        for (int i = 0; i < 225; i++) {
          pixels.setPixelColor(0, pixels.ColorHSV(map(i, 0, 255, 0, 65535)));
          pixels.show();
          delay(currentSweptedDelay);
          if (exitFlag) {
            break;
          }
        }
        clearLED();
        delay(300);
        // Swepted down
        for (int i = 224; i >= 0; i--) {
          pixels.setPixelColor(0, pixels.ColorHSV(map(i, 0, 255, 0, 65535)));
          pixels.show();
          delay(currentSweptedDelay);
          if (exitFlag) {
            break;
          }
        }
        clearLED();
        delay(1000);
      }
      break;
  }
}

// Stepped display pattern
void displaySteppedPattern() {
  switch (currentDir) {
    case 0:
      {
        // Stepped up
        for (int i = 0; i < 8; i++) {
          pixels.setPixelColor(0, pixels.ColorHSV(map(i * COLOR_STEP_INC, 0, 359, 0, 65535)));
          pixels.show();
          delay(currentSteppedDelay);
          if (exitFlag) {
            break;
          }
        }
        clearLED();
        delay(1000);
      }
      break;
    case 1:
      {
        // Stepped up then down
        // Stepped up
        for (int i = 0; i < 8; i++) {
          pixels.setPixelColor(0, pixels.ColorHSV(map(i * COLOR_STEP_INC, 0, 359, 0, 65535)));
          pixels.show();
          delay(currentSteppedDelay);
          if (exitFlag) {
            break;
          }
        }
        clearLED();
        delay(300);
        // Stepped down
        for (int i = 7; i >= 0; i--) {
          pixels.setPixelColor(0, pixels.ColorHSV(map(i * COLOR_STEP_INC, 0, 359, 0, 65535)));
          pixels.show();
          delay(currentSteppedDelay);
          if (exitFlag) {
            break;
          }
        }
        clearLED();
        delay(1000);
      }
      break;
  }
}

// Dispatch display pattern
void dispatchDisplayPattern() {
  exitFlag = false;

  // Process current speed value
  switch (currentSpeed) {
    case 0:
      currentSteppedDelay = SLOW_DELAY_MS;
      currentSweptedDelay = SLOW_DELAY_MS / COLOR_STEP_INC;
      break;
    case 1:
      currentSteppedDelay = MEDIUM_DELAY_MS;
      currentSweptedDelay = MEDIUM_DELAY_MS / COLOR_STEP_INC;
      break;
    case 2:
      currentSteppedDelay = FAST_DELAY_MS;
      currentSweptedDelay = FAST_DELAY_MS / COLOR_STEP_INC;
      break;
  }
  // Process mode value
  switch (currentMode) {
    case 0:
      displaySweptedPattern();
      break;
    case 1:
      displaySteppedPattern();
      break;
  }
}

/*********************************************************************/
/***                          Program Setup                        ***/
/*********************************************************************/

void setup() {

// Initialize the NeoPixel driver
  pixels.begin();
  pixels.clear();
  pixels.show();

  // Initialize Serial port
  Serial.begin(115200);
  delay(1000);
  Serial.println("\n\nStarting Up\n");

  // Bring up WiFi on local network
  WiFi.disconnect();
  WiFi.mode(WIFI_STA);
  WiFi.hostname(WIFI_HOSTNAME);
  WiFi.begin(WIFI_SSID, WIFI_PSWD);
  int i = 0;
  while ((WiFi.status() != WL_CONNECTED) && (i++ < 100)) {
    Serial.print(".");
    delay(500);
  }

  // Did we connect ?
  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("\n\nWiFi connected");
  } else {
    Serial.println("\n\nWiFi did not connect. Restarting ...");
    delay(3000);
    ESP.restart();
  }

  // Set operational defaults
  // Power: 0 = off, 1 = on
  currentPower = 1;

  // Mode: 0 = swept mode, 1 = stepped mode
  currentMode = 0;

  // Dir: 0 = up, 1 = up/down
  currentDir = 0;

  // Speed: 0 = slow, 1 = medium, 2 = fast
  currentSpeed = 1;

  initWebSocket();

  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request) {
    request->send_P(200, "text/html", index_html);
  });

  // Start the server
  server.begin();
}

/*********************************************************************/
/***                           Program Loop                        ***/
/*********************************************************************/

void loop() {

  // Is power on ?
  if (currentPower) {
    dispatchDisplayPattern();
  } else {
    clearLED();
  }

  uint32_t now = millis();

  if (now - last_wifi_check_time > WIFI_TIMEOUT) {
    Serial.print("Checking WiFi... ");
    if (WiFi.status() != WL_CONNECTED) {
      Serial.println("WiFi connection lost. Restarting...\n\n\n");
      delay(3000);
      ESP.restart();
    } else {
      Serial.println("WiFi OK");
    }
    last_wifi_check_time = now;

    // Now check for Internet connectivity
    if (checkInternetConnection() == false) {
      Serial.println("WiFi connection lost. Restarting...\n\n\n");
      delay(3000);
      ESP.restart();
    }
  }
  ws.cleanupClients();
}
